---
title: 'RayFlare: flexible optical modelling of solar cells'

tags:
  - Python
  - physics
  - optics
  - solar cells
  - ray-tracing
  - rigorous coupled-wave analysis
  - transfer matrix method
  - multi-scale modelling
  
authors:
  - name: Phoebe M. Pearce
    orcid: 0000-0001-9082-9506
    affiliation: 1
    
affiliations:
 - name: Department of Physics, University of Cambridge, 19 JJ Thomson Avenue, Cambridge CB3 0HE
   index: 1

date: 26 May 2021
bibliography: paper.bib

---

# Summary

Optical modelling is a key part of the design of solar cells, as the number of absorbed photons in the photovoltaic
layer(s) fundamentally limits the current which can be generated by the cell. Some types of cell design, such as 
multi-junction cells (where multiple semiconductors with different bandgaps are stacked in a single structure to 
each absorb a different part of the solar spectrum) or ultra-thin solar cells [@Hirst2019], require even more careful optical design.
In the first case, it is generally desirable to match the current generated in each layer, requiring tuning of e.g. the
bandgaps and layer thicknesses, while in the second case light-trapping structures such as gratings or surface textures
are required to maintain high absorption and thus current while reducing the semiconductor thickness.

Light-trapping structures range from the relatively simple (e.g. dielectric layers used as anti-reflection coatings
on the front surface of all commercial solar cells) to highly engineered structures for very specific purposes (e.g.
diffraction gratings for ultra-thin gallium arsenide (GaAs) solar cells for space applications). As such, many different simulation methods 
exist which are appropriate for modelling the effect of different light-trapping structures. These include:

- **Simple Beer-Lambert law calculations**: assumes the intensity of light in the material varies as
  $I(z) = I_0 e^{-\alpha z}$, where $z$ is the depth in the material and $\alpha$ is the absorption coefficient. This 
  neglects any effects due to reflection at interfaces, interference, diffraction etc.
- **Ideal/theoretical reference cases**: such as a perfect mirror or perfect Lambertian scatterer. These are often used for comparison
  purposes to benchmark the performance of realistic structures.
- **Ray-tracing (RT)**: a geometric optics method suitable for calculating reflection and refraction due to large-scale 
  (compared to the wavelength) textures, such as the pyramidal texturing commonly using for silicon solar cells.
- **Transfer-matrix method (TMM)**: a wave-optical method suitable for calculating interference effects in thin films (transparent or absorbing).
  This is useful for e.g. calculating the effect of an anti-reflection coating on the front surface of a cell, or
  a distributed Bragg reflector (DBR) designed to reflect specific wavelengths of light.
- **Full Maxwell solvers**: including rigorous coupled-wave analysis (RCWA), also referred to as the Fourier modal method (FMM),
  finite element modelling (FEM) or finite-difference time-domain (FDTD) methods. Since these are more general solutions for
  solving Maxwell's equations in (periodically) textured layers, these have many uses, but are often used in solar cell
  optics research to calculate the effect of gratings or arrays of nanoparticles on absorption in the cell.

![Schematic of the methods implemented in RayFlare. \label{fig:rayflare}](rayflare_diagram_crop.pdf){ width=100% }

# Statement of need

The aim of RayFlare is to provide a single integrated environment with a consistent approach to defining structures which
incorporates all the methods listed above (currently using RCWA as the Maxwell solver), as shown in \autoref{fig:rayflare}. 
The implementation allows users 
to simulate a whole structure using a single method, or to combine different methods and simulate the surfaces of the cell
separately, if appropriate. RayFlare is a further development of the optical modelling capabilities developed for Solcore,
an integrated solar cell modelling package [@Alonso-Alvarez2018].

Existing packages already implement one or more of these methods; for instance, the Python package `tmm` [@Byrnes2016] performs
TMM calculations, while $S^4$ [@Liu2012], written in C++ but including a Python interface, implements RCWA. Many ray-tracing packages are
available such as SunSolve [@SunSolve] which is widely used in solar cell research. There are also existing methods which combine 
multiple methods, such as CROWM [@Lipovsek2011] (ray-tracing with TMM for textured surfaces with thin-film layers), GenPro4 [@Santbergen2017]
and OPTOS [@Eisenlohr2015; @Tucher2015]. The latter two use an angular redistribution matrix method, which allows the
angle-dependent reflection and transmission properties of the front and back surface of the cell to be calculated separately
(further information on this framework is given below).
GenPro4 is a proprietary software which handles matrix calculation for the surfaces using TMM, ray-tracing, or combinations thereof, 
in addition to the subsequent matrix multiplication, while
OPTOS is an open-source code which implements only the matrix multiplication, but not the generation of the matrices describing
the surfaces. Thus, by including TMM, ray-tracing and RCWA methods as well as the angular redistribution matrix method, RayFlare provides
a powerful modelling environment for a very wide range of structures. In addition, the ability to calculate not just
total absorption per layer in the cell but also depth-dependent absorption profiles through the angular redistribution matrix
framework is novel.

# RayFlare

The RT, TMM or RCWA methods can be used independently to simulate a whole structure. In this case, the relevant materials and thicknesses of the layers (all methods), surface textures (RT) or grating shapes
(RCWA) are defined by the user and the reflection, transmission, absorption per layer and (if specified) absorption profile
in the structure are calculated by RayFlare. However, these methods can also be used in combination with each other by using the angular redistribution matrix formalism [@Eisenlohr2015]. In this
framework, the front and rear surface of the structure are decoupled from one another, and suitable methods are used for each surface
to calculate the redistribution matrices. These matrices described how light striking the surface from different directions
is scattered into other directions (which can be in transmission or reflection) or absorbed by any surface layers. The front and
back surface are coupled by some bulk medium, in which simple Beer-Lambert absorption is assumed. This reduces the problem down
to a matrix multiplication process to calculate total reflection, transmission, and absorption in the
surface layers and bulk medium. Note that for the Beer-Lambert absorption approximation to be valid, the bulk medium must be thick
enough that interference effects in it can be neglected, i.e. significantly thicker than the maximum wavelength being considered.

While this matrix multiplication framework has been previously established, as discussed above, the novelty of RayFlare is that
it will automatically calculate the necessary redistribution matrices for the user-defined structure
at the required wavelengths as well as performing the matrix multiplication, including tracking absorption in any surface layers on the
front and rear surface. It can also calculate absorption profiles in the surface and bulk layers, something which has not previously been
demonstrated in combination with the redistribution matrix approach.

The methods currently available to generate redistribution matrices are:

- **Ray-tracing with the Fresnel equations**: suitable for interfaces with some large-scale texture and no surface layers. 
  The ray-tracer was written specifically for RayFlare.
- **TMM**: suitable for planar surfaces with multiple layers. The TMM method implemented in RayFlare is a modified and vectorized
  version of the `tmm` package [@Byrnes2016].
- **Integrated ray-tracing and transfer-matrix method**: the probability of reflection/transmission
  (or absorption in interface layers) are calculated using TMM rather than through the Fresnel equations. This is suitable for large-scale
  textures with one or more thin (compared to the size of the texture features) surface layers.
- **RCWA**: suitable for gratings or other periodic wavelength-scale structures, 
  which can be made of multiple layers including planar layers. RayFlare uses a modified version of $S^4$
  [@Liu2012] which is compatible with Python3 to perform RCWA calculations.
- **Ideal/theoretical reference cases**: currently, a perfect mirror or a perfect Lambertian scatterer are implemented.
  
As each of these methods are implemented independently, additional methods which can be used to calculate redistribution
matrices can be added. 

The results of RayFlare's ray-tracer have been compared with those of PVLighthouse's wafer ray tracer tool [@Waferraytracer],
while the results of the matrix framework using planar surfaces (TMM), a grating (RCWA) or pyramids (ray-tracing) have been
compared to results generated using OPTOS [@Tucher2015], showing excellent agreement in both cases. RayFlare, or code which 
was integrated into RayFlare, has been used to simulate the emissivity/absorptivity
of silicon heterojunction solar cells [@Alonso-Alvarez2019c], design and simulate ultra-thin GaAs solar cells [@Pearce2020],
and to simulate a perovskite on silicon tandem cell [@Pearce2019a]

# Acknowledgements

I would like to thank Diego Alonso Álvarez for helping me innumerable times with Python, package management, and general software developing
skills over the years, and Luis Usier Pimentel for coming up with the name `RayFlare'.

# References
